package im.angry.openeuicc.ui

import android.app.Activity
import android.app.AlertDialog
import android.content.pm.PackageManager
import android.os.Bundle
import android.telephony.SubscriptionManager
import android.telephony.TelephonyManager
import android.telephony.euicc.EuiccManager
import android.text.TextUtils
import android.util.Log

import im.angry.openeuicc.R

class UserConsentActivity : Activity() {

    private lateinit var euiccManager: EuiccManager
    private lateinit var subscriptionManager: SubscriptionManager
    private lateinit var telephonyManager: TelephonyManager
    private var mCardId: Int = -1
    private var mUsePortIndex: Boolean = false
    private var mPortIndex: Int = -1
    private var mSubscriptionId: Int = -1

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        euiccManager = getSystemService(EuiccManager::class.java) as EuiccManager
        subscriptionManager = getSystemService(SubscriptionManager::class.java) as SubscriptionManager
        telephonyManager = getSystemService(TelephonyManager::class.java) as TelephonyManager

        val callingPackage = intent.getStringExtra("android.service.euicc.extra.RESOLUTION_CALLING_PACKAGE")
        if (TextUtils.isEmpty(callingPackage)) {
            Log.e("UserConsentActivity", "Unknown calling package in resolution activity")
            sendResultAndFinish(false)
            return
        }

        val extras = intent.extras
        if (extras != null) {
            mCardId = extras.getInt("android.service.euicc.extra.RESOLUTION_CARD_ID", -1)
            mUsePortIndex = extras.getBoolean("android.service.euicc.extra.RESOLUTION_USE_PORT_INDEX", false)
            mPortIndex = extras.getInt("android.service.euicc.extra.RESOLUTION_PORT_INDEX", -1)
            mSubscriptionId = extras.getInt("android.service.euicc.extra.RESOLUTION_SUBSCRIPTION_ID", -1)
        }

        try {
            val packageManager = packageManager
            val appLabel = packageManager.getApplicationLabel(packageManager.getApplicationInfo(callingPackage!!, 0))
            if (TextUtils.isEmpty(appLabel)) {
                Log.e("UserConsentActivity", "Calling app doesn't have a label")
                sendResultAndFinish(false)
                return
            }
        } catch (e: PackageManager.NameNotFoundException) {
            Log.e("UserConsentActivity", "Calling package doesn't exist", e)
            sendResultAndFinish(false)
            return
        }

        showUserConsentDialog()
    }

    fun sendResultAndFinish(consent: Boolean) {
        val bundle = Bundle()

        bundle.putBoolean("android.service.euicc.extra.RESOLUTION_CONSENT", consent)
        if (mUsePortIndex) {
            bundle.putInt("android.service.euicc.extra.RESOLUTION_PORT_INDEX", mPortIndex);
        }

        euiccManager.createForCardId(mCardId)
        euiccManager.continueOperation(intent, bundle)

        setResult(if (consent) Activity.RESULT_OK else Activity.RESULT_CANCELED)
        finish()
    }

    fun showUserConsentDialog() {
        AlertDialog.Builder(this)
            .setTitle(getString(R.string.user_consent_title))
            .setMessage(getString(R.string.user_consent_message))
            .setPositiveButton(getString(R.string.user_consent_positive)) { _, _ ->
                sendResultAndFinish(true)
            }
            .setNegativeButton(getString(R.string.user_consent_negative)) { _, _ ->
                sendResultAndFinish(false)
            }
            .setOnCancelListener {
                sendResultAndFinish(false)
            }
            .show()
    }
}
